Package: COTAN
Type: Package
Title: COexpression Tables ANalysis
Version: 2.12.0
Authors@R: c(
  person("Galfrè", "Silvia Giulia",email = "silvia.galfre@di.unipi.it", role=c("aut","cre"),comment = c(ORCID = "0000-0002-2770-0344")),
  person("Morandin","Francesco", email = "francesco.morandin@unipr.it", role = "aut",comment = c(ORCID = "0000-0002-2022-2300")),
  person("Fantozzi", "Marco", email = "marco.fantozzi@gmail.com", role = "aut", comment = c(ORCID = "0000-0002-0708-5495")),
  person("Pietrosanto","Marco", role = "aut",comment = c(ORCID = "0000-0001-5129-6065")),
  person("Puttini","Daniel", role = "aut", comment = c(ORCID = "0009-0006-8401-9949")),
  person("Priami","Corrado", role = "aut", comment = c(ORCID = "0000-0002-3261-6235")),
  person("Cremisi","Federico", role = "aut",comment = c(ORCID = "0000-0003-4925-2703")),
  person("Helmer-Citterich","Manuela", role = "aut",comment = c(ORCID = "0000-0001-9530-7504"))
  )
Description: Statistical and computational method to analyze the co-expression of 
    gene pairs at single cell level. It provides the foundation for single-cell gene 
    interactome analysis. The basic idea is studying the zero UMI counts' distribution 
    instead of focusing on positive counts; this is done with a generalized contingency 
    tables framework. COTAN can effectively assess the correlated or anti-correlated 
    expression of gene pairs. It provides a numerical index related to the correlation and an 
    approximate p-value for the associated independence test. COTAN can also evaluate whether 
    single genes are differentially expressed, scoring them with a newly defined global 
    differentiation index. Moreover, this approach provides ways to plot and cluster genes
    according to their co-expression pattern with other genes, effectively helping the study 
    of gene interactions and becoming a new tool to identify cell-identity marker genes.
URL: https://github.com/seriph78/COTAN
BugReports: https://github.com/seriph78/COTAN/issues
Depends: R (>= 4.3)
License: GPL-3
Encoding: UTF-8
RoxygenNote: 7.3.3
Roxygen: list(markdown = TRUE)
Imports: stats, methods, grDevices, Matrix, ggplot2, ggrepel, ggdist,
        ggthemes, graphics, parallel, parallelly, tibble, tidyr, dplyr,
        BiocSingular, parallelDist, ComplexHeatmap, circlize, grid,
        scales, RColorBrewer, utils, rlang, Rfast, stringr, Seurat,
        dendextend, zeallot, assertthat, withr, SummarizedExperiment,
        SingleCellExperiment, proxy, RSpectra
Suggests: testthat (>= 3.2.0), proto, spelling, knitr, ragg, Cairo,
        conflicted, data.table, gsubfn, R.utils, tidyverse, rmarkdown,
        htmlwidgets, MASS, Rtsne, plotly, BiocStyle, cowplot, qpdf,
        GEOquery, sf, torch, S4Vectors
Config/testthat/edition: 3
Language: en-US
biocViews: SystemsBiology, Transcriptomics, GeneExpression, SingleCell,
        DifferentialExpression, Clustering, GPU
VignetteBuilder: knitr
LazyData: false
git_url: https://git.bioconductor.org/packages/COTAN
git_branch: RELEASE_3_23
git_last_commit: c7bfb93
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:35:52 UTC; biocbuild
Author: Galfrè Silvia Giulia [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2770-0344>),
  Morandin Francesco [aut] (ORCID:
    <https://orcid.org/0000-0002-2022-2300>),
  Fantozzi Marco [aut] (ORCID: <https://orcid.org/0000-0002-0708-5495>),
  Pietrosanto Marco [aut] (ORCID:
    <https://orcid.org/0000-0001-5129-6065>),
  Puttini Daniel [aut] (ORCID: <https://orcid.org/0009-0006-8401-9949>),
  Priami Corrado [aut] (ORCID: <https://orcid.org/0000-0002-3261-6235>),
  Cremisi Federico [aut] (ORCID: <https://orcid.org/0000-0003-4925-2703>),
  Helmer-Citterich Manuela [aut] (ORCID:
    <https://orcid.org/0000-0001-9530-7504>)
Maintainer: Galfrè Silvia Giulia <silvia.galfre@di.unipi.it>
