% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{Datasets}
\alias{Datasets}
\alias{raw.dataset}
\alias{ERCCraw}
\alias{test.dataset}
\alias{test.dataset.clusters1}
\alias{test.dataset.clusters2}
\alias{vignette.cells.origin}
\alias{vignette.split.clusters}
\alias{vignette.merge.clusters}
\title{Data-sets}
\format{
\code{raw.dataset} is a data frame with \eqn{2000} genes and \eqn{815}
cells

\code{ERCCraw} is a \code{data.frame}

\code{test.dataset} is a \code{data.frame} with \eqn{600} genes and \eqn{1200}
cells

\code{test.dataset.clusters1} is a \verb{character array}

\code{test.dataset.clusters2} is a \verb{character array}

\code{vignette.cells.origin} is a \code{factor}

\code{vignette.split.clusters} is a \code{factor}

\code{vignette.merge.clusters} is a \code{factor}
}
\source{
\code{GEO} \code{GSM2861514}

\code{ERCC}
}
\usage{
data(raw.dataset)

data(ERCCraw)

data(test.dataset)

data(test.dataset.clusters1)

data(test.dataset.clusters2)

data(vignette.cells.origin)

data(vignette.split.clusters)

data(vignette.merge.clusters)
}
\value{
Objects loaded by \code{data()}:
\itemize{
\item \code{raw.dataset}, \code{ERCCraw}, \code{test.dataset}: \code{data.frame}s
\item \code{test.dataset.clusters1}, \code{test.dataset.clusters2}:
character vectors of cluster labels
\item \code{vignette.split.clusters}, \code{vignette.merge.clusters},
\code{vignette.cells.origin}: factors of cluster labels
}
}
\description{
Simple data-sets included in the package
}
\details{
\code{raw.dataset} is a sub-sample of a real \emph{scRNA-seq} data-set

\code{ERCCraw} dataset

\code{test.dataset} is an artificial data set obtained by sampling target
negative binomial distributions on a set of \eqn{600} genes on \eqn{2} two
cells \emph{clusters} of \eqn{600} cells each. Each \emph{clusters} has its own set
of parameters for the distributions even, but a fraction of the genes has
the same expression in both \emph{clusters}.

\code{test.dataset.clusters1} is the \emph{clusterization} obtained running
\code{cellsUniformClustering()} on the \code{test.dataset}

\code{test.dataset.clusters2} is the \emph{clusterization} obtained running
\code{mergeUniformCellsClusters()} on the \code{test.dataset} using the previous
\emph{clusterization}

\code{vignette.cells.origin} is the factor discriminating between the
\emph{cortical} cells and the \emph{other} cells. This was constructed from the
following dataset: Mouse Cortex E17.5, \code{GEO}: \code{GSM2861514}
\url{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSM2861514}

\code{vignette.split.clusters} is the clusterization obtained running
\code{cellsUniformClustering()} on the vignette dataset (mouse cortex E17.5,
\code{GEO}: \code{GSM2861514})

\code{vignette.merge.clusters} is the clusterization obtained running
\code{mergeUniformCellsClusters()} on the vignette dataset (mouse cortex E17.5,
\code{GEO}: \code{GSM2861514}) starting from the \code{vignette.split.clusters}
\emph{clusterization}, with a sequence of progressively relaxed \emph{UT} checks
}
\keyword{datasets}
