% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getColorsVector}
\alias{getColorsVector}
\title{\code{getColorsVector}}
\usage{
getColorsVector(numNeededColors = 0L)
}
\arguments{
\item{numNeededColors}{The number of returned colors. If omitted it returns
all available colors}
}
\value{
an array of \code{RGB} colors of the wanted size
}
\description{
This function returns a list of colors based on the
\code{\link[RColorBrewer:ColorBrewer]{RColorBrewer::brewer.pal()}} function
}
\details{
The colors are taken from the \code{\link[RColorBrewer:ColorBrewer]{RColorBrewer::brewer.pal.info()}} sets
with \code{Set1}, \code{Set2}, \code{Set3} placed first.
}
\examples{
colorsVector <- getColorsVector(17)

}
