% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{COTAN-class}
\alias{COTAN-class}
\title{The \code{COTAN} class}
\value{
A \code{\linkS4class{COTAN}} object.
}
\description{
Definition of the \code{COTAN} class
}
\section{Slots}{

\describe{
\item{\code{raw}}{\code{dgCMatrix} - the raw \code{UMI} count matrix \eqn{n \times m} (gene
number × cell number)}

\item{\code{genesCoex}}{\code{dspMatrix} - the correlation of \code{COTAN} between genes,
\eqn{n \times n}}

\item{\code{cellsCoex}}{\code{dspMatrix} - the correlation of \code{COTAN} between cells,
\eqn{m \times m}}

\item{\code{metaDataset}}{\code{data.frame}}

\item{\code{metaCells}}{\code{data.frame}}

\item{\code{clustersCoex}}{a \code{list} of \code{COEX} \code{data.frames} for each clustering in
the \code{metaCells}}
}}

